/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

//3 dimensions of threads blocks (X,Y,Z) and 2 dimensions of blocks grids (X,Y)
#define X_BLOCK 16
#define Y_BLOCK 1
#define Z_BLOCK 1
#define X_GRID  64
#define Y_GRID  1 

//number of threads in blocks
const int BLOCK_SIZE = X_BLOCK*Y_BLOCK*Z_BLOCK;

//total number of threads in launch
const int NUMBER_THREADS = X_BLOCK*Y_BLOCK*Z_BLOCK*X_GRID*Y_GRID;

#define USE_NAF
#define OWNAF
#define NAF3

#include <iostream>
#include <fstream>
#include "point_multiplication.cuh"

using namespace std;
//string representing the data

//the prime
string prime_str  = "114804434916157216670268400907125902210939926123376911736891363408990795293183";
//parameter a of the curve
string a_str = "40436044618658097711785492504012338403285403687372493464186014894083987988766";
//parameter b of the curve
string b_str = "34886244569985231445566545442256685214700455632045658520045215600554112234535";

//useless : a*R % prime (with R = 2^N)
string aR_str = "35818727692110112246791219884337495308805502403405326499582077407610210068615";

//x coordinate of the base point (affine - jacobian)
string x_str = "98498615158468461235135876454535135434853124353445315312353454331532132186";
//y coordinate of the base point (affine - jacobian)
string y_str = "30504020161418492107222612569116259853304322480488846940224848403070111276424";
//z coordinate of the base point (jacobian)
string z_str = "1";

//useless (x*R % prime, y*R % prime, z*R % prime)
string xR_str= "11621810404147645572992006334957427616308718412382077366295476033029868053677";
string yR_str= "95745984883152444421279714097632615676426113333802962083583322613535555384499";
string zR_str= "987654321158978753302584101562005642330058542263652302566220598922334346753";

//multiplier of the point
string k_str = "73222198712380060011778887725006639011079894034799419387461336330972926146576";

//number of bits
const uint N = 140; 

int main() {

	cudaSetDevice(1);

	//ONE k, ONE curve, ONE point...
	dim3 dim_block(X_BLOCK, Y_BLOCK, Z_BLOCK);
	dim3 dim_grid(X_GRID, Y_GRID);

  
	// define the integer k for [k]P 
	Integer<N> k(k_str); 
	
	// define the curve		
	Integer<N> a (a_str);
	Integer<N> prime(prime_str);
	uint invp=Utils::modulo_inverse((1 << 16), -prime[0]);
 
	// define the point P	
	Integer<N> Px (x_str);
	Integer<N> Py (y_str);
	Integer<N> Pz (z_str);

	//thread vector of integers with k
	vector<Integer<N> > v_k(NUMBER_THREADS,k); 
	/*
	ifstream file("k.txt");
	string line;
	int i = 0;
	while(!file.eof()) {
		getline(file, line);
		if(line !="") {
			v_k[i] = (Integer<N>(line));
		}
		i++;
	}
	file.close();
	*/

	//thread vector of elliptic curve parameters (a,prime, prime^(-1) mod 2^16)
	vector<Integer<N> > v_a(NUMBER_THREADS, a);
	vector<Integer<N> > v_prime(NUMBER_THREADS, prime);
	vector<uint>        v_invp(NUMBER_THREADS, invp);

	//thread vector of point P
	vector<Integer<N> > v_Px(NUMBER_THREADS,Px);
	vector<Integer<N> > v_Py(NUMBER_THREADS,Py);
	vector<Integer<N> > v_Pz(NUMBER_THREADS,Pz);

 
	//thread vector of point [k]P
	vector<Integer<N> > v_kPx(NUMBER_THREADS);
	vector<Integer<N> > v_kPy(NUMBER_THREADS);
	vector<Integer<N> > v_kPz(NUMBER_THREADS); 


	// define the data on GPU 
	GPU_vector_data<N> gpu_data; 
	
	// allocate space on the GPU for [k]P results
	Memory<N, NUMBER_THREADS>::allocGPU(gpu_data.__v_Qx);
	Memory<N, NUMBER_THREADS>::allocGPU(gpu_data.__v_Qy);
	Memory<N, NUMBER_THREADS>::allocGPU(gpu_data.__v_Qz);

	// send the data to the GPU
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_k, v_k);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_a, v_a);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_prime, v_prime);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_invp, v_invp);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Px, v_Px);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Py, v_Py);
	Memory<N, NUMBER_THREADS>::sendToGPU(gpu_data.__v_Pz, v_Pz);
	
	cudaEvent_t start, stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	cudaEventRecord(start, 0); 

	point_multiplication_naf<N><<<dim_grid,dim_block>>>(gpu_data);

	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
	float elapsedTime;
	cudaEventElapsedTime(&elapsedTime, start, stop);
	cudaEventDestroy(start);
	cudaEventDestroy(stop);

	//copy result from device to host
	Memory<N, NUMBER_THREADS>::sendToCPU(v_kPx, gpu_data.__v_Qx);
	Memory<N, NUMBER_THREADS>::sendToCPU(v_kPy, gpu_data.__v_Qy);
	Memory<N, NUMBER_THREADS>::sendToCPU(v_kPz, gpu_data.__v_Qz);
	
	//display result (in jacobian form)
	//	for(uint i = 0; i < NUMBER_THREADS; i++) {
	for(uint i = 0; i < 1; i++) {
		cout << "*************************************************************************************************" << endl;
		cout << "RESULTS : " << i << endl;
		cout << "X_result = " << v_kPx[i].to_string() << endl;
		cout << "Y_result = " << v_kPy[i].to_string() << endl;
		cout << "Z_result = " << v_kPz[i].to_string() << endl;
		cout << "*************************************************************************************************" << endl;
	}
	//TODO : convert in affine (x = X/Z^2, y = Y/Z^3)*/
	cout <<"Total time: "<< elapsedTime <<" ms"<< endl;
	cout <<"Time per [k]P: "<< elapsedTime/NUMBER_THREADS <<" ms"<< endl;

	return 0;
}
