/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
#ifndef UTIL_HPP
#define UTIL_HPP

// define macro for host and device qualifier
#define _HD_ __host__ __device__ 
#define _H_  __host__
#define _D_  __device__

// define macro for unsigned int
typedef unsigned int uint;

class Utils {
public:
	static _HD_ void extended_euclidian(uint op1, uint op2, int& u, int& v, int& gcd);
	static _HD_ int modulo_inverse(uint op1, uint op2);
	static _HD_ int log2ceil(int N);
	static _D_  int tid();
	static _D_  int tid_block();
};

_HD_ void Utils::extended_euclidian(uint op1, uint op2, int& u, int& v, int& gcd) {
	int b = op2;
	gcd = op1;
	int x = 0;
	u = 1;
	int y = 1;
	v = 0;
	while (b != 0) {
		int q = gcd / b;
		int t = b;
		b = gcd % b;
		gcd = t; 
		t = x;
		x = u-q*x;
		u = t;
		t = y;
		y = v-q*y;
		v = t;
	}
}

//gcd(op1,op2) = 1
_HD_ int Utils::modulo_inverse(uint op1, uint op2) {
	int b = op2;
	int a = op1;
	int y = 1;
	int lasty = 0;
	while (b != 0) {
		int q = a / b;
		int t = b;
		b = a % b;
		a = t;
		t = y;
		y = lasty-q*y;
		lasty = t;
	}
	return  lasty;
}

_HD_ int Utils::log2ceil(int N) {
	return ceil(log2((float)N));
}

//return identifier of thread
__device__ int Utils::tid() {
	return threadIdx.x+blockDim.x*threadIdx.y+threadIdx.z*blockDim.x*blockDim.y+blockIdx.x*blockDim.x*blockDim.y*blockDim.z+gridDim.x*blockDim.x*blockDim.y*blockDim.z*blockIdx.y;
}

//return identifier of the thread in the block
__device__ int Utils::tid_block() {
	return threadIdx.x+blockDim.x*threadIdx.y+threadIdx.z*blockDim.x*blockDim.y;
}

#endif
